<?php

namespace Drupal\views_editable_entity\Plugin\views\area;

use Drupal\Component\Serialization\Json;
use Drupal\Component\Utility\Html;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\ContentEntityTypeInterface;
use Drupal\Core\Entity\EntityDisplayRepositoryInterface;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Routing\RedirectDestinationTrait;
use Drupal\views\Attribute\ViewsArea;
use Drupal\views\Plugin\views\area\TokenizeAreaPluginBase;
use Drupal\views\Plugin\views\display\DisplayPluginBase;
use Drupal\views\ViewExecutable;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides an area handler which renders an entity in a certain view mode.
 *
 * @ingroup views_area_handlers
 */
#[ViewsArea("views_editable_entity")]
class EditableEntity extends TokenizeAreaPluginBase {
  use RedirectDestinationTrait;

  /**
   * Stores the entity type of the result entities.
   *
   * @var string
   */
  protected $entityType;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity repository service.
   *
   * @var \Drupal\Core\Entity\EntityRepositoryInterface
   */
  protected $entityRepository;

  /**
   * The entity display repository.
   *
   * @var \Drupal\Core\Entity\EntityDisplayRepositoryInterface
   */
  protected $entityDisplayRepository;

  /**
   * Constructs a new Entity instance.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Entity\EntityRepositoryInterface $entity_repository
   *   The entity repository.
   * @param \Drupal\Core\Entity\EntityDisplayRepositoryInterface $entity_display_repository
   *   The entity display repository.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, EntityTypeManagerInterface $entity_type_manager, EntityRepositoryInterface $entity_repository, EntityDisplayRepositoryInterface $entity_display_repository) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->entityTypeManager = $entity_type_manager;
    $this->entityRepository = $entity_repository;
    $this->entityDisplayRepository = $entity_display_repository;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('entity.repository'),
      $container->get('entity_display.repository')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function init(ViewExecutable $view, DisplayPluginBase $display, ?array &$options = NULL) {
    // Use the view's entity type as a fallback.
    $this->definition['entity_type'] = $view->getBaseEntityType()->id();
    $this->entityType = $this->definition['entity_type'];
    parent::init($view, $display, $options);
  }

  /**
   * {@inheritdoc}
   */
  protected function defineOptions() {
    $options = parent::defineOptions();

    // Per default we enable tokenize, as this is the most common use case for
    // this handler.
    $options['tokenize']['default'] = FALSE;

    // Contains the config target identifier for the entity.
    $options['entity_type'] = ['default' => ''];
    $options['target'] = ['default' => ''];
    $options['button_text'] = ['default' => 'Edit Header'];
    $options['class'] = ['default' => NULL];
    $options['display'] = ['default' => ''];
    $options['width'] = ['default' => '600'];
    $options['view_mode'] = ['default' => 'default'];
    $options['bypass_access'] = ['default' => FALSE];

    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function buildOptionsForm(&$form, FormStateInterface $form_state) {
    $options = $form_state->getValue('options');
    if (!empty($options['entity_type'])) {
      $this->definition['entity_type'] = $options['entity_type'];
    }
    $entity_types = $this->entityTypeManager->getDefinitions();
    $content_entity_types = array_filter($entity_types,
      function ($entity_type) {
        return $entity_type instanceof ContentEntityTypeInterface;
      }
    );
    $type_options = [];
    foreach ($content_entity_types as $type_id => $content_entity_type) {
      $type_options[$type_id] = $content_entity_type->getLabel();
    }

    $entity_type = $this->options['entity_type'] ?: $this->definition['entity_type'] ?: $this->entityType;

    $form['entity_type'] = [
      '#type' => 'select',
      '#options' => $type_options,
      '#title' => $this->t('Entity type'),
      '#default_value' => $entity_type,
      '#ajax' => [
        'callback' => [static::class, 'ajaxRefresh'],
        'disable-refocus' => FALSE,
        'event' => 'change',
        'wrapper' => 'editable-entity--target',
        'progress' => [
          'type' => 'throbber',
          'message' => $this->t('Updating...'),
        ],
      ],
    ];

    $target = $this->options['target'];

    // If the target does not contain tokens, try to load the entity and
    // display the entity ID to the admin form user.
    // @todo Use a method to check for tokens in
    //   https://www.drupal.org/node/2396607.
    if (!str_contains($this->options['target'], '{{')) {
      // @todo If the entity does not exist, this will show the config target
      //   identifier. Decide if this is the correct behavior in
      //   https://www.drupal.org/node/2415391.
      if (!empty($this->options['target']) && $target_entity = $this->entityRepository->loadEntityByConfigTarget($entity_type, $this->options['target'])) {
        $target = $target_entity;
      }
      else {
        $target = NULL;
      }
    }

    // If there's a value submitted for the select list set the textfield value.
    if (!$target_type = $options['entity_type'] ?? '') {
      $target_type = $this->options['entity_type'] ?: $this->definition['entity_type'];
    }

    if ($target instanceof ContentEntityInterface && $target->getEntityTypeId() !== $target_type) {
      $target = NULL;
    }

    $form['target'] = [
      '#type' => 'entity_autocomplete',
      '#title' => t('Entity'),
      '#default_value' => $target,
      '#target_type' => $target_type,
      '#validate_reference' => FALSE,
      '#prefix' => '<div id="editable-entity--target">',
      '#suffix' => '</div>',
      '#required' => TRUE,
    ];

    $form['view_mode'] = [
      '#type' => 'select',
      '#options' => $this->entityDisplayRepository->getViewModeOptions($this->entityType),
      '#title' => $this->t('View mode'),
      '#default_value' => $this->options['view_mode'],
    ];

    $form['button_text'] = [
      '#title' => $this->t('Edit button text'),
      '#type' => 'textfield',
      '#default_value' => $this->options['button_text'] ?? '',
      '#required' => TRUE,
    ];

    $form['class'] = [
      '#title' => $this->t('Class'),
      '#description' => $this->t('A CSS class to apply to the link. If using multiple classes, separate them by spaces.'),
      '#type' => 'textfield',
      '#default_value' => $this->options['class'],
    ];

    $form['display'] = [
      '#title' => $this->t('How to display'),
      '#description' => $this->t('Optionally have the form open on-page in a modal or off-canvas dialog.'),
      '#type' => 'select',
      '#default_value' => $this->options['display'],
      '#options' => [
        '' => $this->t('Default'),
        'tray' => $this->t('Off-Screen Tray'),
        'modal' => $this->t('Modal Dialog'),
      ],
    ];

    $form['width'] = [
      '#title' => $this->t('Dialog Width'),
      '#description' => $this->t('How wide the dialog should appear.'),
      '#type' => 'number',
      '#min' => '100',
      '#default_value' => $this->options['width'],
      '#states' => [
        // Show this number field only if a dialog is chosen above.
        'invisible' => [
          ':input[name="options[display]"]' => ['value' => ''],
        ],
      ],
    ];

    $form['bypass_access'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Bypass access checks'),
      '#description' => $this->t('If enabled, access permissions for rendering the entity are not checked.'),
      '#default_value' => !empty($this->options['bypass_access']),
    ];
  }

  /**
   * Refresh the autocomplete target field based on an entity type change.
   *
   * @param array $form
   *   The form definition.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form values.
   *
   * @return array
   *   The updated form element.
   */
  public static function ajaxRefresh(array &$form, FormStateInterface $form_state): array {
    // Return the prepared autocomplete, with any value set to null.
    $form['options']['target']['#value'] = NULL;
    return $form['options']['target'];
  }

  /**
   * {@inheritdoc}
   */
  public function submitOptionsForm(&$form, FormStateInterface $form_state) {
    $options = $form_state->getValue('options');
    $this->entityType = $options['entity_type'];
    $this->definition['entity_type'] = $options['entity_type'];
    parent::submitOptionsForm($form, $form_state);

    // Load the referenced entity and store its config target identifier if
    // the target does not contains tokens.
    // @todo Use a method to check for tokens in
    //   https://www.drupal.org/node/2396607.
    if (!empty($options['entity_type'])) {
      $this->definition['entity_type'] = $options['entity_type'];
    }
    if (!str_contains($options['target'], '{{')) {
      if ($entity = $this->entityFromOptions($options['target'], $options['entity_type'])) {
        $options['target'] = $entity->getConfigTarget();
        $form_state->setValue('options', $options);
      }
    }
  }

  /**
   * Normalize the target to an entity object.
   *
   * @param mixed $target
   *   The entity that should be referenced.
   * @param string $target_type
   *   The entity type to references.
   *
   * @return null|ContentEntityInterface
   *   A normalized entity object, or null.
   */
  public function entityFromOptions(mixed $target, string $target_type): null|ContentEntityInterface {
    if (is_string($target) && strlen($target) === 36) {
      return $this->entityRepository->loadEntityByUuid($target_type, $target);
    }
    if ($target instanceof ContentEntityInterface) {
      return $target;
    }
    if ($entity = $this->entityTypeManager->getStorage($target_type)->load($target)) {
      return $entity;
    }
    return NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function render($empty = FALSE) {
    $return = [];
    // If we don't know the entity type, nothing to show.
    if (!$entity_type = $this->options['entity_type']) {
      return [];
    }
    if (!$empty || !empty($this->options['empty'])) {

      // @todo Use a method to check for tokens in
      //   https://www.drupal.org/node/2396607.
      if (str_contains($this->options['target'], '{{')) {
        // We cast as we need the integer/string value provided by the
        // ::tokenizeValue() call.
        $target_id = (string) $this->tokenizeValue($this->options['target']);
        if ($entity = $this->entityTypeManager->getStorage($entity_type)->load($target_id)) {
          $target_entity = $entity;
        }
      }
      else {
        if ($entity = $this->entityRepository->loadEntityByConfigTarget($entity_type, $this->options['target'])) {
          $target_entity = $entity;
        }
      }
      if (isset($target_entity) && (!empty($this->options['bypass_access']) || $target_entity->access('view'))) {
        $params = $this->getDestinationArray();
        $actions = [];
        if ($target_entity->access('edit')) {
          // The entity can be edited.
          $edit_txt = $this->options['button_text'] ?? $this->t('Edit');
          $element = $target_entity->toLink($edit_txt, 'edit-form', ['query' => $params])->toRenderable();

          // Parse and sanitize provided classes.
          if ($this->options['class']) {
            $classes = explode(' ', $this->options['class']);
            foreach ($classes as $class) {
              $element['#options']['attributes']['class'][] = Html::getClass($class);
            }
          }

          // Apply the selected dialog options.
          if ($this->options['display']) {
            $element['#options']['attributes']['class'][] = 'use-ajax';
            $width = $this->options['width'] ?: 600;
            $element['#options']['attributes']['data-dialog-options'] = Json::encode(['width' => $width]);
            switch ($this->options['display']) {
              case 'tray':
                $element['#options']['attributes']['data-dialog-renderer'] = 'off_canvas';
                $element['#options']['attributes']['data-dialog-type'] = 'dialog';
                break;

              case 'modal':
                $element['#options']['attributes']['data-dialog-type'] = 'modal';
                break;
            }
          }
          $actions['edit'] = $element;
        }
        if ($actions) {
          $return['actions'] = $actions;
        }
        $view_builder = $this->entityTypeManager->getViewBuilder($entity_type);
        $return['entity'] = $view_builder->view($target_entity, $this->options['view_mode']);
      }
    }

    return $return;
  }

  /**
   * {@inheritdoc}
   */
  public function calculateDependencies() {
    $dependencies = parent::calculateDependencies();
    $entity_type = $this->options['entity_type'] ?? $this->entityType;

    // Ensure that we don't add dependencies for placeholders.
    // @todo Use a method to check for tokens in
    //   https://www.drupal.org/node/2396607.
    if (!str_contains($this->options['target'], '{{')) {
      if ($entity = $this->entityRepository->loadEntityByConfigTarget($entity_type, $this->options['target'])) {
        $dependencies[$this->entityTypeManager->getDefinition($entity_type)->getConfigDependencyKey()][] = $entity->getConfigDependencyName();
      }
    }

    return $dependencies;
  }

}
